#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>
#include <stdbool.h>
#include <ncurses.h>
#include <unistd.h>
#include <termios.h>
#include <limits.h>
#include <time.h>
#include <SDL2/SDL.h>
#include <SDL2/SDL_image.h>

#include "hash.h"
#include "structure.h"
#include "base.h"
#include "move.h"
#include "generation.h"
#include "display.h"

int* drawOrder;

void init_draworder() {
    drawOrder = malloc(sizeof(int)*6) ;
}

void updateRenderer(SDL_Renderer* renderer) {
    //printf("E");
    SDL_RenderPresent(renderer);
}

void resetRenderer(SDL_Renderer* renderer) {
    SDL_SetRenderDrawColor(renderer, 0, 0, 0, 255);
    SDL_RenderClear(renderer);
}

void drawRectToRenderer(SDL_Renderer* renderer, SDL_Rect* rect, int R, int G, int B, int A) {
    SDL_SetRenderDrawColor(renderer, R, G, B, A);
    SDL_RenderFillRect(renderer, rect);
}

void placeRectToRenderer(SDL_Renderer* renderer, int X, int Y, int W, int H, int R, int G, int B, int A) {
    SDL_Rect rect;
    rect.x = X;
    rect.y = Y;
    rect.w = W;
    rect.h = H;
    SDL_SetRenderDrawColor(renderer, R, G, B, A);
    SDL_RenderFillRect(renderer, &rect);
}

void placeRectToRendererNoColor(SDL_Renderer* renderer, int X, int Y, int W, int H) {
    SDL_Rect rect;
    rect.x = X;
    rect.y = Y;
    rect.w = W;
    rect.h = H;
    SDL_RenderFillRect(renderer, &rect);
}

void drawLineWithThicc(SDL_Renderer* renderer, int width, int x1, int x2, int y1, int y2, int R, int G, int B, int A) {
    // draws line with width (native SDL cannot do that)
    double theta = 0.0;
    double seglen = distance_pt_pt_3d(x1, y1, 0.0, x2, y2, 0.0);
    while(theta < 1.0) {
        placeRectToRenderer(renderer, convex_seg(x1, x2, theta)-width/2, convex_seg(y1, y2, theta)-width/2, width, width, R, G, B, A);
        theta += 1 / seglen;
    }
}

void drawLineWithThiccNoColor(SDL_Renderer* renderer, int width, int x1, int x2, int y1, int y2) {
    // draws line with width (native SDL cannot do that)
    double theta = 0.0;
    double seglen = distance_pt_pt_3d(x1, y1, 0.0, x2, y2, 0.0);
    while(theta < 1.0) {
        placeRectToRendererNoColor(renderer, convex_seg(x1, x2, theta)-width/2, convex_seg(y1, y2, theta)-width/2, width, width);
        theta += 1 / seglen;
    }
}

void drawLineWithThiccGradient(SDL_Renderer* renderer, int start_width, int end_width, int x1, int x2, int y1, int y2, int R, int G, int B, int A) {
    // draws line with width ;
    // width goes from start_width to end_with in a linear way
    double theta = 0.0;
    double seglen = distance_pt_pt_3d(x1, y1, 0.0, x2, y2, 0.0);
    while(theta < 1.0) {
        int width = convex_seg(start_width, end_width, theta) ;
        placeRectToRenderer(renderer, convex_seg(x1, x2, theta)-width/2, convex_seg(y1, y2, theta)-width/2, width, width, R, G, B, A);
        theta += 1 / seglen;
    }
}

void drawDigitToRenderer(SDL_Renderer* renderer, imgs data, int digit, int X, int Y, int W, int H) {
    if(digit == -727) {
        SDL_Rect rect;
        rect.x = X;
        rect.y = Y;
        rect.w = W;
        rect.h = H;
            
        SDL_Texture* texture = data.arr[10];
        
        SDL_RenderCopy(renderer, texture, NULL, &rect);
    } else if (!(0 <= digit && digit <= 9)) {
        fprintf(stderr, "Illegal digit : '%d'.\n", digit);
        exit(1);
    } else {
        SDL_Rect rect;
        rect.x = X;
        rect.y = Y;
        rect.w = W;
        rect.h = H;
            
        SDL_Texture* texture = data.arr[digit];
        
        SDL_RenderCopy(renderer, texture, NULL, &rect);
    }
}

void drawCharToRenderer(SDL_Renderer* renderer, imgs data, char c, int X, int Y, int W, int H) {
    if ((int)c >= 97 && (int)c <= 122) {
        SDL_Rect rect;
        rect.x = X;
        rect.y = Y;
        rect.w = W;
        rect.h = H;
            
        SDL_Texture* texture = data.arr[(int)c - 97];
        
        SDL_RenderCopy(renderer, texture, NULL, &rect);
    }
}

void drawNumberToRenderer(SDL_Renderer* renderer, imgs data, int n, int X, int Y, int W, int H, int Woffset) {
    if(n == 0) {
        drawDigitToRenderer(renderer, data, 0, X + W, Y, W, H);
    } else if(n > 0) {
        int toDraw = 0, remaining = n, nIter = ln_baseN(n, 10);
        while(nIter != 0) {
            toDraw = remaining%10;
            remaining = remaining / 10;
            drawDigitToRenderer(renderer, data, toDraw, X + (W-Woffset)*nIter, Y, W, H);
            nIter--;
        }
    } else {
        int toDraw = 0, remaining = -n, nIter = ln_baseN(-n, 10);
        drawDigitToRenderer(renderer, data, -727, X, Y, W, H);
        while(nIter != 0) {
            toDraw = remaining%10;
            remaining = remaining / 10;
            drawDigitToRenderer(renderer, data, toDraw, X + (W-Woffset)*nIter, Y, W, H);
            nIter--;
        }
    }
}

void drawStringToRenderer(SDL_Renderer* renderer, imgs data, char* s, int X, int Y, int W, int H) {
    int k = 0 ;
    while(s[k] != '\0') {
        drawCharToRenderer(renderer, data, s[k], X + W*k, Y, W, H);
        k += 1;
    }
}

// ---------------------------------------------------------------------------------------------------------------------------------- //
// 3D stuff

void project_to_camera(double x0, double y0, double z0, double* rx, double* ry, double* rz) {
    // align pt to (0, 0, 0)
    double x = x0 - camx ;
    double y = y0 - camy ;
    double z = z0 - camz ;

    // rotate (y)
    double xry = x*cos(rot_hz) + z*sin(rot_hz) ;
    double yry = y ;
    double zry = z*cos(rot_hz) - x*sin(rot_hz) ;

    // rotate (x)
    *rx = xry ;
    *ry = yry*cos(rot_vt) - zry*sin(rot_vt) ;
    *rz = zry*cos(rot_vt) + yry*sin(rot_vt) ;
}

void rotate_cube(double x0, double y0, double z0, double* rx, double* ry, double* rz, cube_0 cb) {
    // align pt to (0, 0, 0)
    double x = x0 - (cb.x + cb.w/2) ;
    double y = y0 - (cb.y + cb.h/2) ;
    double z = z0 - (cb.z + cb.d/2) ;

    // rotate (y)
    double xry = x*cos(cb.hz_angle) - z*sin(cb.hz_angle) ;
    double yry = y ;
    double zry = z*cos(cb.hz_angle) + x*sin(cb.hz_angle) ;

    // rotate (x)
    *rx = (cb.x + cb.w/2) + xry ;
    *ry = (cb.y + cb.h/2) + yry*cos(cb.vt_angle) - zry*sin(cb.vt_angle) ;
    *rz = (cb.z + cb.d/2) + zry*cos(cb.vt_angle) + yry*sin(cb.vt_angle) ;
}

void draw_segment(SDL_Renderer* renderer, double sx, double sy, double sz, double ex, double ey, double ez) {
    double spx ;double spy ;double spz ;double epx ;double epy ;double epz ;
    project_to_camera(sx, sy, sz, &spx, &spy, &spz) ;
    project_to_camera(ex, ey, ez, &epx, &epy, &epz) ;
    if(spz >= 0.4 && epz >= 0.4) {
        drawLineWithThiccNoColor(renderer, 1, 
            (int)(1500.0 * (1.0 + (spx / (1.5 * spz * tan_fov))) / 2.0),
            (int)(1500.0 * (1.0 + (epx / (1.5 * epz * tan_fov))) / 2.0),
            (int)(1000.0 * (1.0 + (spy / (spz * tan_fov))) / 2.0),
            (int)(1000.0 * (1.0 + (epy / (epz * tan_fov))) / 2.0)
        ) ;
    } else if(epz >= 0.4) {
        double midx = convex_pt(ex, sx, (epz - 0.4)/(epz - spz)) ;
        double midy = convex_pt(ey, sy, (epz - 0.4)/(epz - spz)) ;
        double midz = convex_pt(ez, sz, (epz - 0.4)/(epz - spz)) ;
        project_to_camera(midx, midy, midz, &spx, &spy, &spz) ;
        //printf("* %f\n", spz) ;
        drawLineWithThiccNoColor(renderer, 1, 
            (int)(1500.0 * (1.0 + (spx / (1.5 * spz * tan_fov))) / 2.0),
            (int)(1500.0 * (1.0 + (epx / (1.5 * epz * tan_fov))) / 2.0),
            (int)(1000.0 * (1.0 + (spy / (spz * tan_fov))) / 2.0),
            (int)(1000.0 * (1.0 + (epy / (epz * tan_fov))) / 2.0)
        ) ;
    } else if(spz >= 0.4) {
        double midx = convex_pt(sx, ex, (spz - 0.4)/(spz - epz)) ;
        double midy = convex_pt(sy, ey, (spz - 0.4)/(spz - epz)) ;
        double midz = convex_pt(sz, ez, (spz - 0.4)/(spz - epz)) ;
        project_to_camera(midx, midy, midz, &epx, &epy, &epz) ;
        //printf("%f *\n", epz) ;
        drawLineWithThiccNoColor(renderer, 1, 
            (int)(1500.0 * (1.0 + (spx / (1.5 * spz * tan_fov))) / 2.0),
            (int)(1500.0 * (1.0 + (epx / (1.5 * epz * tan_fov))) / 2.0),
            (int)(1000.0 * (1.0 + (spy / (spz * tan_fov))) / 2.0),
            (int)(1000.0 * (1.0 + (epy / (epz * tan_fov))) / 2.0)
        ) ;
    } // else : not visible (behind camera)
}

void axialRotation_X0(double* y, double* z, double theta) {
    double y0 = *y ;
    *y = (*y)*cos(theta) - (*z)*sin(theta) ;
    *z = (*z)*cos(theta) + y0*sin(theta) ;
}

void axialRotation_X(double* y, double* z, double theta, double cst_y, double cst_z) {
    // project the space onto the (y, z) plane to get cst_y and cst_z
    double y1 = *y - cst_y;
    double z1 = *z - cst_z;
    axialRotation_X0(&y1, &z1, theta);
    *y = y1 + cst_y ;
    *z = z1 + cst_z ;
}

void axialRotation_Y0(double* x, double* z, double theta) {
    double x0 = *x ;
    *x = (*x)*cos(theta) + (*z)*sin(theta) ;
    *z = (*z)*cos(theta) - x0*sin(theta) ;
}

void axialRotation_Y(double* x, double* z, double theta, double cst_x, double cst_z) {
    // project the space onto the (y, z) plane to get cst_y and cst_z
    double x1 = *x - cst_x;
    double z1 = *z - cst_z;
    axialRotation_Y0(&x1, &z1, theta);
    *x = x1 + cst_x ;
    *z = z1 + cst_z ;
}

void axialRotation_Z0(double* x, double* y, double theta) {
    double x0 = *x ;
    *x = (*x)*cos(theta) - (*y)*sin(theta) ;
    *y = (*y)*cos(theta) + x0*sin(theta) ;
}

void axialRotation_Z(double* x, double* y, double theta, double cst_x, double cst_y) {
    // project the space onto the (y, z) plane to get cst_y and cst_z
    double x1 = *x - cst_x;
    double y1 = *y - cst_y;
    axialRotation_Z0(&x1, &y1, theta);
    *x = x1 + cst_x ;
    *y = y1 + cst_y ;
}

void drawSegmentRotated(SDL_Renderer* renderer, double sx, double sy, double sz, double ex, double ey, double ez, double hz_angle, double vt_angle, double center_x, double center_y, double center_z) {
    double psx = sx;
    double psy = sy;
    double psz = sz;
    double pex = ex;
    double pey = ey;
    double pez = ez;
    axialRotation_Y(&psx, &psz, -hz_angle, center_x, center_z);
    axialRotation_Y(&pex, &pez, -hz_angle, center_x, center_z);
    // minus signs cause the base is indirect
    axialRotation_X(&psy, &psz, vt_angle, center_y, center_z);
    axialRotation_X(&pey, &pez, vt_angle, center_y, center_z);
    draw_segment(renderer, psx, psy, psz, pex, pey, pez);
}

void drawOutlineOfCube_0(SDL_Renderer* renderer, cube_0 c) {
    // x = constant
    drawSegmentRotated(renderer, c.x, c.y, c.z, c.x + c.w, c.y, c.z, c.hz_angle, c.vt_angle, c.x + c.w/2.0, c.y + c.h/2.0, c.z + c.d/2.0) ;
    drawSegmentRotated(renderer, c.x, c.y + c.h, c.z, c.x + c.w, c.y + c.h, c.z, c.hz_angle, c.vt_angle, c.x + c.w/2.0, c.y + c.h/2.0, c.z + c.d/2.0) ;
    drawSegmentRotated(renderer, c.x, c.y, c.z + c.d, c.x + c.w, c.y, c.z + c.d, c.hz_angle, c.vt_angle, c.x + c.w/2.0, c.y + c.h/2.0, c.z + c.d/2.0) ;
    drawSegmentRotated(renderer, c.x, c.y + c.h, c.z + c.d, c.x + c.w, c.y + c.h, c.z + c.d, c.hz_angle, c.vt_angle, c.x + c.w/2.0, c.y + c.h/2.0, c.z + c.d/2.0) ;

    // y = constant
    drawSegmentRotated(renderer, c.x, c.y, c.z, c.x, c.y + c.h, c.z, c.hz_angle, c.vt_angle, c.x + c.w/2.0, c.y + c.h/2.0, c.z + c.d/2.0) ;
    drawSegmentRotated(renderer, c.x + c.w, c.y, c.z, c.x + c.w, c.y + c.h, c.z, c.hz_angle, c.vt_angle, c.x + c.w/2.0, c.y + c.h/2.0, c.z + c.d/2.0) ;
    drawSegmentRotated(renderer, c.x, c.y, c.z + c.d, c.x, c.y + c.h, c.z + c.d, c.hz_angle, c.vt_angle, c.x + c.w/2.0, c.y + c.h/2.0, c.z + c.d/2.0) ;
    drawSegmentRotated(renderer, c.x + c.w, c.y, c.z + c.d, c.x + c.w, c.y + c.h, c.z + c.d, c.hz_angle, c.vt_angle, c.x + c.w/2.0, c.y + c.h/2.0, c.z + c.d/2.0) ;

    // z = constant
    drawSegmentRotated(renderer, c.x, c.y, c.z, c.x, c.y, c.z + c.d, c.hz_angle, c.vt_angle, c.x + c.w/2.0, c.y + c.h/2.0, c.z + c.d/2.0) ;
    drawSegmentRotated(renderer, c.x + c.w, c.y, c.z, c.x + c.w, c.y, c.z + c.d, c.hz_angle, c.vt_angle, c.x + c.w/2.0, c.y + c.h/2.0, c.z + c.d/2.0) ;
    drawSegmentRotated(renderer, c.x, c.y + c.h, c.z, c.x, c.y + c.h, c.z + c.d, c.hz_angle, c.vt_angle, c.x + c.w/2.0, c.y + c.h/2.0, c.z + c.d/2.0) ;
    drawSegmentRotated(renderer, c.x + c.w, c.y + c.h, c.z, c.x + c.w, c.y + c.h, c.z + c.d, c.hz_angle, c.vt_angle, c.x + c.w/2.0, c.y + c.h/2.0, c.z + c.d/2.0) ;
}

// -------------------------------------------------------------------------------------------------------------------------------- //
// this is where the fun begins

int surfaceDrawOrder(double x0, double y0, double z0, cube_0 cb) {
    // returns the number of surfaces that should be drawn, as well as filling drawOrder for said surfaces :
    // 0 = +x ; 1 = -x
    // 2 = +y ; 3 = -y
    // 4 = +z ; 5 = -z

    // align cube center to (0, 0, 0)
    double x = x0 - (cb.x + cb.w/2.0) ;
    double y = y0 - (cb.y + cb.h/2.0) ;
    double z = z0 - (cb.z + cb.d/2.0) ;
    // rotate (y)
    double xry = x*cos(cb.hz_angle) + z*sin(cb.hz_angle) ;
    double yry = y ;
    double zry = z*cos(cb.hz_angle) - x*sin(cb.hz_angle) ;
    // rotate (x)
    double xrx = xry ;
    double yrx = yry*cos(cb.vt_angle) - zry*sin(cb.vt_angle) ;
    double zrx = zry*cos(cb.vt_angle) + yry*sin(cb.vt_angle) ;
    // cube is centered and aligned
    int id = 0 ;
    if(xrx > cb.w/2.0) {
        drawOrder[id] = 0 ;
        id += 1 ;
    } else if(xrx < -cb.w/2.0) {
        drawOrder[id] = 1 ;
        id += 1 ;
    }
    if(yrx > cb.h/2.0) {
        drawOrder[id] = 2 ;
        id += 1 ;
    } else if(yrx < -cb.h/2.0) {
        drawOrder[id] = 3 ;
        id += 1 ;
    }
    if(zrx > cb.d/2.0) {
        drawOrder[id] = 4 ;
        id += 1 ;
    } else if(zrx < -cb.d/2.0) {
        drawOrder[id] = 5 ;
        id += 1 ;
    }
    if(id == 0) { // inside the cube
        for(int k = 0; k < 6; k++) {
            drawOrder[k] = k ;
        }
        return 6;
    } else {
        return id ;
    }
}

SDL_Vertex construct_vertex(double px, double py, int r, int g, int b) {
    SDL_Vertex vtx ;
    vtx.color.r = r ;
    vtx.color.g = g ;
    vtx.color.b = b ;
    vtx.color.a = SDL_ALPHA_OPAQUE;
    vtx.position.x = (float)px ;
    vtx.position.y = (float)py ;
    vtx.tex_coord.x = 0.0f;
    vtx.tex_coord.y = 0.0f;
    return vtx ;
}

void renderTriangle(
    SDL_Renderer* renderer,
    double x0, double y0, double z0,
    double x1, double y1, double z1,
    double x2, double y2, double z2,
    int red, int green, int blue
) {
    double px0; double py0; double pz0;
    double px1; double py1; double pz1;
    double px2; double py2; double pz2;
    project_to_camera(x0, y0, z0, &px0, &py0, &pz0);
    project_to_camera(x1, y1, z1, &px1, &py1, &pz1);
    project_to_camera(x2, y2, z2, &px2, &py2, &pz2);
    if(pz0 >= 0.4 && pz1 >= 0.4 && pz2 >= 0.4) {
        const SDL_Vertex vtxs[3] = {
            construct_vertex(1500.0 * (1.0 + (px0 / (1.5 * pz0 * tan_fov))) / 2.0, 1000.0 * (1.0 + (py0 / (pz0 * tan_fov))) / 2.0, red, green, blue),
            construct_vertex(1500.0 * (1.0 + (px1 / (1.5 * pz1 * tan_fov))) / 2.0, 1000.0 * (1.0 + (py1 / (pz1 * tan_fov))) / 2.0, red, green, blue),
            construct_vertex(1500.0 * (1.0 + (px2 / (1.5 * pz2 * tan_fov))) / 2.0, 1000.0 * (1.0 + (py2 / (pz2 * tan_fov))) / 2.0, red, green, blue),
        };
        SDL_RenderGeometry(renderer, NULL, vtxs, 3, NULL, 0);
    } else if((pz0 >= 0.4) + (pz1 >= 0.4) + (pz2 >= 0.4) == 2) {
        double fpx0; double fpy0; double fpz0;
        double fpx1; double fpy1; double fpz1;
        double mpx0; double mpy0; double mpz0;
        double mpx1; double mpy1; double mpz1;
        if(pz0 < 0.4) {     
            // pz1 >= 0.4 and pz2 >+ 0.4
            fpx0 = px1 ; fpy0 = py1 ; fpz0 = pz1 ;
            fpx1 = px2 ; fpy1 = py2 ; fpz1 = pz2 ;
            // 1-0 segment
            project_to_camera(
                convex_pt(x1, x0, (pz1 - 0.4)/(pz1 - pz0)), 
                convex_pt(y1, y0, (pz1 - 0.4)/(pz1 - pz0)), 
                convex_pt(z1, z0, (pz1 - 0.4)/(pz1 - pz0)), 
            &mpx0, &mpy0, &mpz0) ;
            
            // 0-2 segment
            project_to_camera(
                convex_pt(x2, x0, (pz2 - 0.4)/(pz2 - pz0)), 
                convex_pt(y2, y0, (pz2 - 0.4)/(pz2 - pz0)), 
                convex_pt(z2, z0, (pz2 - 0.4)/(pz2 - pz0)), 
            &mpx1, &mpy1, &mpz1) ;
        } else if(pz1 < 0.4) {     
            // pz0 >= 0.4 and pz2 >+ 0.4
            fpx0 = px0 ; fpy0 = py0 ; fpz0 = pz0 ;
            fpx1 = px2 ; fpy1 = py2 ; fpz1 = pz2 ;
            // 0-1 segment
            project_to_camera(
                convex_pt(x0, x1, (pz0 - 0.4)/(pz0 - pz1)), 
                convex_pt(y0, y1, (pz0 - 0.4)/(pz0 - pz1)), 
                convex_pt(z0, z1, (pz0 - 0.4)/(pz0 - pz1)), 
            &mpx0, &mpy0, &mpz0) ;
            
            // 1-2 segment
            project_to_camera(
                convex_pt(x2, x1, (pz2 - 0.4)/(pz2 - pz1)), 
                convex_pt(y2, y1, (pz2 - 0.4)/(pz2 - pz1)), 
                convex_pt(z2, z1, (pz2 - 0.4)/(pz2 - pz1)), 
            &mpx1, &mpy1, &mpz1) ;
        } else if(pz2 < 0.4) {     
            // pz1 >= 0.4 and pz0 >+ 0.4
            fpx0 = px0 ; fpy0 = py0 ; fpz0 = pz0 ;
            fpx1 = px1 ; fpy1 = py1 ; fpz1 = pz1 ;
            // 0-2 segment
            project_to_camera(
                convex_pt(x0, x2, (pz0 - 0.4)/(pz0 - pz2)), 
                convex_pt(y0, y2, (pz0 - 0.4)/(pz0 - pz2)), 
                convex_pt(z0, z2, (pz0 - 0.4)/(pz0 - pz2)), 
            &mpx0, &mpy0, &mpz0) ;
            
            // 1-2 segment
            project_to_camera(
                convex_pt(x1, x2, (pz1 - 0.4)/(pz1 - pz2)), 
                convex_pt(y1, y2, (pz1 - 0.4)/(pz1 - pz2)), 
                convex_pt(z1, z2, (pz1 - 0.4)/(pz1 - pz2)), 
            &mpx1, &mpy1, &mpz1) ;
        }

        const SDL_Vertex vtxs[6] = {
            construct_vertex(1500.0 * (1.0 + (fpx0 / (1.5 * fpz0 * tan_fov))) / 2.0, 1000.0 * (1.0 + (fpy0 / (fpz0 * tan_fov))) / 2.0, red, green, blue),
            construct_vertex(1500.0 * (1.0 + (mpx0 / (1.5 * mpz0 * tan_fov))) / 2.0, 1000.0 * (1.0 + (mpy0 / (mpz0 * tan_fov))) / 2.0, red, green, blue),
            construct_vertex(1500.0 * (1.0 + (fpx1 / (1.5 * fpz1 * tan_fov))) / 2.0, 1000.0 * (1.0 + (fpy1 / (fpz1 * tan_fov))) / 2.0, red, green, blue),
            construct_vertex(1500.0 * (1.0 + (mpx0 / (1.5 * mpz0 * tan_fov))) / 2.0, 1000.0 * (1.0 + (mpy0 / (mpz0 * tan_fov))) / 2.0, red, green, blue),
            construct_vertex(1500.0 * (1.0 + (mpx1 / (1.5 * mpz1 * tan_fov))) / 2.0, 1000.0 * (1.0 + (mpy1 / (mpz1 * tan_fov))) / 2.0, red, green, blue),
            construct_vertex(1500.0 * (1.0 + (fpx1 / (1.5 * fpz1 * tan_fov))) / 2.0, 1000.0 * (1.0 + (fpy1 / (fpz1 * tan_fov))) / 2.0, red, green, blue),
        };
        SDL_RenderGeometry(renderer, NULL, vtxs, 6, NULL, 0);
    } else if((pz0 >= 0.4) + (pz1 >= 0.4) + (pz2 >= 0.4) == 1) {
        if(pz0 >= 0.4) {
            project_to_camera(
                convex_pt(x0, x1, (pz0 - 0.4)/(pz0 - pz1)),
                convex_pt(y0, y1, (pz0 - 0.4)/(pz0 - pz1)),
                convex_pt(z0, z1, (pz0 - 0.4)/(pz0 - pz1)),
            &px1, &py1, &pz1);
            project_to_camera(
                convex_pt(x0, x2, (pz0 - 0.4)/(pz0 - pz2)),
                convex_pt(y0, y2, (pz0 - 0.4)/(pz0 - pz2)),
                convex_pt(z0, z2, (pz0 - 0.4)/(pz0 - pz2)),
            &px2, &py2, &pz2);
        } else if(pz1 >= 0.4) {
            project_to_camera(
                convex_pt(x1, x0, (pz1 - 0.4)/(pz1 - pz0)),
                convex_pt(y1, y0, (pz1 - 0.4)/(pz1 - pz0)),
                convex_pt(z1, z0, (pz1 - 0.4)/(pz1 - pz0)),
            &px0, &py0, &pz0);
            project_to_camera(
                convex_pt(x1, x2, (pz1 - 0.4)/(pz1 - pz2)),
                convex_pt(y1, y2, (pz1 - 0.4)/(pz1 - pz2)),
                convex_pt(z1, z2, (pz1 - 0.4)/(pz1 - pz2)),
            &px2, &py2, &pz2);
        } else if(pz2 >= 0.4) {
            project_to_camera(
                convex_pt(x2, x0, (pz2 - 0.4)/(pz2 - pz0)),
                convex_pt(y2, y0, (pz2 - 0.4)/(pz2 - pz0)),
                convex_pt(z2, z0, (pz2 - 0.4)/(pz2 - pz0)),
            &px0, &py0, &pz0);
            project_to_camera(
                convex_pt(x2, x1, (pz2 - 0.4)/(pz2 - pz1)),
                convex_pt(y2, y1, (pz2 - 0.4)/(pz2 - pz1)),
                convex_pt(z2, z1, (pz2 - 0.4)/(pz2 - pz1)),
            &px1, &py1, &pz1);
        }
        const SDL_Vertex vtxs[3] = {
            construct_vertex(1500.0 * (1.0 + (px0 / (1.5 * pz0 * tan_fov))) / 2.0, 1000.0 * (1.0 + (py0 / (pz0 * tan_fov))) / 2.0, red, green, blue),
            construct_vertex(1500.0 * (1.0 + (px1 / (1.5 * pz1 * tan_fov))) / 2.0, 1000.0 * (1.0 + (py1 / (pz1 * tan_fov))) / 2.0, red, green, blue),
            construct_vertex(1500.0 * (1.0 + (px2 / (1.5 * pz2 * tan_fov))) / 2.0, 1000.0 * (1.0 + (py2 / (pz2 * tan_fov))) / 2.0, red, green, blue),
        };
        SDL_RenderGeometry(renderer, NULL, vtxs, 3, NULL, 0);
    }
}

void renderTriangleRotated(
    SDL_Renderer* renderer,
    double x0, double y0, double z0,
    double x1, double y1, double z1,
    double x2, double y2, double z2,
    int red, int green, int blue,
    cube_0 cb
) {
    double px0; double py0; double pz0;
    double px1; double py1; double pz1;
    double px2; double py2; double pz2;
    rotate_cube(x0, y0, z0, &px0, &py0, &pz0, cb);
    rotate_cube(x1, y1, z1, &px1, &py1, &pz1, cb);
    rotate_cube(x2, y2, z2, &px2, &py2, &pz2, cb);
    renderTriangle(renderer, px0, py0, pz0, px1, py1, pz1, px2, py2, pz2, red, green, blue);
}

void drawSfOfCube(SDL_Renderer* renderer, int sf, cube_0 c) {
    if(sf == 0 || sf == 1) {        // x
        renderTriangleRotated(renderer, 
            c.x + c.w*(sf==0), c.y, c.z, 
            c.x + c.w*(sf==0), c.y + c.h, c.z, 
            c.x + c.w*(sf==0), c.y + c.h, c.z + c.d, 
            c.red, c.green, c.blue, c
        );
        renderTriangleRotated(renderer, 
            c.x + c.w*(sf==0), c.y, c.z, 
            c.x + c.w*(sf==0), c.y, c.z + c.d, 
            c.x + c.w*(sf==0), c.y + c.h, c.z + c.d, 
            c.red, c.green, c.blue, c
        );
    } else if(sf == 2 || sf == 3) { // y
        renderTriangleRotated(renderer, 
            c.x, c.y + c.h*(sf==2), c.z, 
            c.x + c.w, c.y + c.h*(sf==2), c.z, 
            c.x + c.w, c.y + c.h*(sf==2), c.z + c.d, 
            c.red, c.green, c.blue, c
        );
        renderTriangleRotated(renderer, 
            c.x, c.y + c.h*(sf==2), c.z, 
            c.x, c.y + c.h*(sf==2), c.z + c.d, 
            c.x + c.w, c.y + c.h*(sf==2), c.z + c.d, 
            c.red, c.green, c.blue, c
        );
    } else {                        // z
        renderTriangleRotated(renderer, 
            c.x, c.y, c.z + c.d*(sf==4), 
            c.x + c.w, c.y, c.z + c.d*(sf==4), 
            c.x + c.w, c.y + c.h, c.z + c.d*(sf==4), 
            c.red, c.green, c.blue, c
        );
        renderTriangleRotated(renderer, 
            c.x, c.y, c.z + c.d*(sf==4), 
            c.x, c.y + c.h, c.z + c.d*(sf==4), 
            c.x + c.w, c.y + c.h, c.z + c.d*(sf==4), 
            c.red, c.green, c.blue, c
        );
    }
}

void drawFullCube(SDL_Renderer* renderer, cube_0 cb) {
    int sfToDraw = surfaceDrawOrder(camx, camy, camz, cb);
    for(int k = 0; k < sfToDraw; k++) {
        drawSfOfCube(renderer, drawOrder[k], cb);
    }
}

// -------------------------------------------------------------------------------------------------------------------------------- //

void swap_cb(cube_0* arr, int i, int j) {
    cube_0 temp = arr[i];
    arr[i] = arr[j];
    arr[j] = temp;
}

void swap_tp(teleporter* arr, int i, int j) {
    teleporter temp = arr[i];
    arr[i] = arr[j];
    arr[j] = temp;
}

void insertionSort_cb(cube_0* arr, int len) {
    for(int k = 0; k < len; k++) {
        int j = k-1 ;
        while(j >= 0) {
            if(distance_pt_cube_0_3d(camx, camy, camz, arr[j]) < distance_pt_cube_0_3d(camx, camy, camz, arr[j+1])) {
                swap_cb(arr, j, j+1);
                j -= 1;
            } else {
                j = -1;
            }
        }
    }
}

void insertionSort_tp(teleporter* arr, int len) {
    for(int k = 0; k < len; k++) {
        int j = k-1 ;
        while(j >= 0) {
            if(distance_pt_cube_0_3d(camx, camy, camz, arr[j].hitbox) < distance_pt_cube_0_3d(camx, camy, camz, arr[j+1].hitbox)) {
                swap_tp(arr, j, j+1);
                j -= 1;
            } else {
                j = -1;
            }
        }
    }
}

void drawCurrentRoom(SDL_Renderer* renderer) {
    insertionSort_cb(current_room->map, current_room->map_size);
    insertionSort_tp(current_room->tps, current_room->tps_size);
    for(int k = 0; k < current_room->map_size; k++) {
        SDL_SetRenderDrawColor(renderer, current_room->map[k].red/2, current_room->map[k].green/2, current_room->map[k].blue/2, 255);
        drawOutlineOfCube_0(renderer, current_room->map[k]);
        
        drawFullCube(renderer, current_room->map[k]);
    }
    for(int k = 0; k < current_room->tps_size; k++) {
        SDL_SetRenderDrawColor(renderer, current_room->tps[k].hitbox.red/2,  current_room->tps[k].hitbox.green/2,  current_room->tps[k].hitbox.blue/2, 255);
        drawOutlineOfCube_0(renderer, current_room->tps[k].hitbox);
        
        drawFullCube(renderer, current_room->tps[k].hitbox);
    }
}

// -------------------------------------------------------------------------------------------------------------------------------- //

void drawData(SDL_Renderer* renderer) {
    drawNumberToRenderer(renderer, digits, (int)camx, 10, 10, 75/2, 105/2, 0) ;
    drawNumberToRenderer(renderer, digits, (int)camy, 10, 60, 75/2, 105/2, 0) ;
    drawNumberToRenderer(renderer, digits, (int)camz, 10, 110, 75/2, 105/2, 0) ;

    drawNumberToRenderer(renderer, digits, (int)(rot_hz*180.0/3.14159), 10, 190, 75/2, 105/2, 0) ;
    drawNumberToRenderer(renderer, digits, (int)(rot_vt*180.0/3.14159), 10, 240, 75/2, 105/2, 0) ;

    drawNumberToRenderer(renderer, digits, player_chx, 310, 10, 75/2, 105/2, 0);
    drawNumberToRenderer(renderer, digits, player_chy, 310, 60, 75/2, 105/2, 0);
}